import createPub from '../';
import {usePub}  from '../';

export type QueryFuncType = (...args: Array<any>) => Promise;

export type CreateQueryPubOptions = {
    queryFunc?: QueryFuncType,
    cacheTimeout?: number,
    hidePendingOnRefresh?: boolean
    avoidRepeatInProgress?: boolean,
    forceRefreshAfterTimeout?: boolean,
}

enum PromiseStatus  {
    PENDING = "PENDING",
    FULFILLED = "FULFILLED",
    REJECTED = "REJECTED"
}

type PromiseSnapshot = {
    status: PromiseStatus,
    payload: any,
    error: Error
}

const defaultOptions:CreateQueryPubOptions = {
    cacheTimeout: 2 * 60 * 1000,
    hidePendingOnRefresh: true,
    avoidRepeatInProgress: true,
    forceRefreshAfterTimeout: true,
}

export default (options:CreateQueryPubOptions) => {
    const queryPub = createPub({queries: {}, mutations: {}})
    const currentPromises = {queries: {}, mutations: {}}

    options.cacheTimeout                ||= defaultOptions.cacheTimeout
    options.hidePendingOnRefresh        ||= defaultOptions.hidePendingOnRefresh
    options.avoidRepeatInProgress       ||= defaultOptions.avoidRepeatInProgress
    options.forceRefreshAfterTimeout    ||= defaultOptions.forceRefreshAfterTimeout

    /*
     * function to remember basic transport function (like fetch or gql) in closure for further usage
     *
     * @param {function} queryFunc? - basic transport function. If omitted, used function from options of createQueryPub
     * @returns {function} "createQuery", which receives promise name for pub, isMutation flag, cacheTagsFunc and query function arguments 
     */
    const withQueryFunc = (queryFunc?: QueryFuncType):Function => {
        //TODO: give ability of options override
        return ({promiseName, name, isMutation, cacheTagsFunc, args=[]}) => {
            promiseName = promiseName || name

            const qF = queryFunc || options.queryFunc
            const defaultArgs = args
            

            /*
             * Concats default arguments from withQueryFunc resulting function with passed arguments
             * 
             */
            const mixArgs = (args:any[]) => 
                Array.isArray(defaultArgs) ? 
                    [...defaultArgs, ...args] :
                    (typeof defaultArgs === 'function' ? 
                        defaultArgs(...args) : args)

            /*
             * used to create key for pub
             */
            const stringifyArgs = args => 
                JSON.stringify(mixArgs(args))

            /*
             * invokes query, using basic transport function and mixArgs
             */
            const makeQuery = (...args:any[]) => {
                return qF(...mixArgs(args))
            }

            /*
             * Checks tags intersection for invalidation of some cached query result
             */
            const intersectTags = (first:Array<any>, second:Array<any>) => 
                first.find(firstItem => second.find(secondItem => firstItem.type === secondItem.type && firstItem.id === secondItem.id))


            const invalidateTags = tags => {
                console.log(tags)
                if (!tags) return
                for (const promiseBranch of Object.values(queryPub.queries)){
                    for (const [args, promiseState] of Object.entries(promiseBranch))
                        if (promiseState.cacheTags && intersectTags(tags, promiseState.cacheTags)){
                            console.log('invalidate', tags, args, promiseState)
                            delete promiseBranch[args]
                        }
                }
            }

            /*
             * Performs query, if needed, using makeQuery or forceQuery
             * Checks is query mutation or pure query, and options like avoidRepeatInProgress or 
             * cache timeouts. So, in some circumstances, query skipped.
             */
            const query = (...args:any[]) => {
                if (isMutation){
                    if (options.avoidRepeatInProgress && queryPub.mutations[promiseName]?.refreshing){
                        return currentPromises.mutations[promiseName]
                    }
                    else {
                        if (!options.hidePendingOnRefresh || !queryPub.mutations[promiseName]){
                            queryPub.mutations[promiseName] = {status: 'PENDING', refreshing: true}
                        }
                        if (!queryPub.mutations[promiseName].refreshing){
                            queryPub.mutations[promiseName].refreshing = true
                        }
                        return currentPromises.mutations[promiseName] = makeQuery(...args).then(payload => queryPub.mutations[promiseName] = (invalidateTags(cacheTagsFunc && cacheTagsFunc(payload)),{status: 'FULFILLED', payload}),
                                                                                                error   => queryPub.mutations[promiseName] = {status: 'REJECTED', error})
                    }
                }
                if (!isMutation){
                    const stringifiedArgs = stringifyArgs(args)
                    if ((!options.cacheTimeout && queryPub.queries[promiseName]?.[stringifiedArgs].status === 'FULFILLED') || 
                        (queryPub.queries[promiseName]?.[stringifiedArgs] && (Date.now() - queryPub.queries[promiseName][stringifiedArgs].timeStamp) < options.cacheTimeout)){

                        return currentPromises.queries[promiseName][stringifiedArgs] 
                    }
                    return forceQuery(...args)
                }
            }

            /*
             * Forces query using makeQuery, but maintain pub structure. (cache timeout, status and so)
             * Always runs query for Mutation
             * The only reason to skip query for Query is avoidRepeatInProgress (deduplicate queries). Maybe this logic should be moved to query function
             *
             */
            const forceQuery = (...args:any[]) =>  {
                if (isMutation) return query(...args)



                const stringifiedArgs = stringifyArgs(args)

                if (!queryPub.queries[promiseName]) {
                    queryPub.queries[promiseName] = {}
                }
                if (!currentPromises.queries[promiseName]){
                    currentPromises.queries[promiseName] = {}
                }


                if (options.avoidRepeatInProgress && queryPub.queries[promiseName]?.[stringifiedArgs]?.refreshing){
                    return currentPromises.queries[promiseName][stringifiedArgs]
                }

                if (!options.hidePendingOnRefresh || !queryPub.queries[promiseName]?.[stringifiedArgs]){
                    queryPub.queries[promiseName][stringifiedArgs] = {status: 'PENDING', timeStamp: Date.now(), refreshing: true}
                }

                if (!queryPub.queries[promiseName][stringifiedArgs].refreshing){
                    queryPub.queries[promiseName][stringifiedArgs].refreshing = true
                }

                if (queryPub.queries[promiseName][stringifiedArgs].timeoutId){
                    clearTimeout(queryPub.queries[promiseName][stringifiedArgs].timeoutId)
                }



                return currentPromises.queries[promiseName][stringifiedArgs] = makeQuery(...args)
                                                                                    .then(payload => queryPub.queries[promiseName][stringifiedArgs] = {    
                                                                                                                                                            status: 'FULFILLED', 
                                                                                                                                                            payload, 
                                                                                                                                                            timeStamp: Date.now(), 
                                                                                                                                                            cacheTags: cacheTagsFunc && cacheTagsFunc(payload),
                                                                                                                                                            timeoutId: options.cacheTimeout && options.forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), options.cacheTimeout) : null
                                                                                                                                                        },
                                                                                           error   => queryPub.queries[promiseName][stringifiedArgs] = {
                                                                                                                                                            status: 'REJECTED', 
                                                                                                                                                            error, 
                                                                                                                                                            timeStamp: Date.now(),
                                                                                                                                                            timeoutId: options.cacheTimeout && options.forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), options.cacheTimeout) : null
                                                                                                                                                        })
            }

            /*
             * Creates hook name from promise name like rtk query
             */
            const hookName = `use${promiseName[0].toUpperCase() + promiseName.slice(1) + (isMutation ? 'Mutation' : 'Query')}`

            /*
             * Mutation hook. Returns query function and current pub data
             */
            const useMutation = () => {
                usePub(queryPub.mutations)

                return [query, queryPub.mutations[promiseName] || {status: 'INIT'}]
            }

            /*
             * Query hook. Calls query function with arguments passed. 
             * Listening for pub.
             * @returns pub data
             */
            const useQuery = (...args) => {
                query(...args)

                usePub(queryPub.queries[promiseName])

                const stringifiedArgs = stringifyArgs(args)
                return queryPub.queries[promiseName][stringifiedArgs]
            }

            /*
             * Gets sub pub for this promiseName and this parameters. Used by NamedRoute
             * to subscribe on pub and refresh when pub updates. Working only for Query,
             * not for Mutations
             */
            query.getQueryPubBranch = (...args) => {
                const stringifiedArgs = stringifyArgs(args)
                return [queryPub.queries[promiseName], stringifiedArgs] 
            }

            return {makeQuery, query, forceQuery, [hookName]: isMutation ? useMutation : useQuery, hook: isMutation ? useMutation : useQuery, hookName}
        }
    }

    return {queryPub, withQueryFunc, ...(options.queryFunc ? {createQuery: withQueryFunc(options.queryFunc)}: {})}
}
